@@ -17,14 +17,18 @@
 
 #include <QByteArray>
 #include <QDateTime>
+#include <QHash>
 #include <QImage>
 #include <QIODevice>
+#include <QLoggingCategory>
 #include <QPoint>
 #include <QSize>
 #include <QSharedPointer>
 
 #include "microexif_p.h"
 
+Q_DECLARE_LOGGING_CATEGORY(LOG_IFFPLUGIN)
+
 // Main chunks (Standard)
 #define CAT__CHUNK QByteArray("CAT ")
 #define FILL_CHUNK QByteArray("    ")
@@ -49,10 +53,16 @@
 #define BODY_CHUNK QByteArray("BODY")
 #define CAMG_CHUNK QByteArray("CAMG")
 #define CMAP_CHUNK QByteArray("CMAP")
+#define CMYK_CHUNK QByteArray("CMYK") // https://wiki.amigaos.net/wiki/ILBM_IFF_Interleaved_Bitmap#ILBM.CMYK
 #define DPI__CHUNK QByteArray("DPI ")
+#define XBMI_CHUNK QByteArray("XBMI")
 
-#define CTBL_CHUNK QByteArray("CTBL") // undocumented
-#define SHAM_CHUNK QByteArray("SHAM") // undocumented
+// Different palette for scanline
+#define BEAM_CHUNK QByteArray("BEAM")
+#define CTBL_CHUNK QByteArray("CTBL") // same as BEAM
+#define PCHG_CHUNK QByteArray("PCHG") // encoded in a unknown way (to be investigated)
+#define RAST_CHUNK QByteArray("RAST") // Atari ST(E)
+#define SHAM_CHUNK QByteArray("SHAM")
 
 // FOR4 CIMG IFF (Maya)
 #define RGBA_CHUNK QByteArray("RGBA")
@@ -64,6 +74,7 @@
 #define COPY_CHUNK QByteArray("(c) ")
 #define DATE_CHUNK QByteArray("DATE")
 #define EXIF_CHUNK QByteArray("EXIF") // https://aminet.net/package/docs/misc/IFF-metadata
+#define ICCN_CHUNK QByteArray("ICCN") // https://aminet.net/package/docs/misc/IFF-metadata
 #define ICCP_CHUNK QByteArray("ICCP") // https://aminet.net/package/docs/misc/IFF-metadata
 #define FVER_CHUNK QByteArray("FVER")
 #define HIST_CHUNK QByteArray("HIST")
@@ -74,16 +85,28 @@
 #define ACBM_FORM_TYPE QByteArray("ACBM")
 #define ILBM_FORM_TYPE QByteArray("ILBM")
 #define PBM__FORM_TYPE QByteArray("PBM ")
+#define RGB8_FORM_TYPE QByteArray("RGB8")
+#define RGBN_FORM_TYPE QByteArray("RGBN")
+
 #define CIMG_FOR4_TYPE QByteArray("CIMG")
 #define TBMP_FOR4_TYPE QByteArray("TBMP")
 
 #define CHUNKID_DEFINE(a) static QByteArray defaultChunkId() { return a; }
 
+// The 8-bit RGB format must be consistent. If you change it here, you have also to use the same
+// when converting an image with BEAM/CTBL/SHAM chunks otherwise the option(QImageIOHandler::ImageFormat)
+// could returns a wrong value.
+// Warning: Changing it requires changing the algorithms. Se, don't touch! :)
+#define FORMAT_RGB_8BIT QImage::Format_RGB888 // default one
+
+#define FORMAT_RGBA_8BIT QImage::Format_RGBA8888 // used by PCHG chunk
+
 /*!
  * \brief The IFFChunk class
  */
 class IFFChunk
 {
+    friend class IFFHandlerPrivate;
 public:
     using ChunkList = QList<QSharedPointer<IFFChunk>>;
 
@@ -300,18 +323,30 @@ protected:
     inline quint16 ui16(quint8 c1, quint8 c2) const {
         return (quint16(c2) << 8) | quint16(c1);
     }
+    inline quint16 ui16(const QByteArray &data, qint32 pos) const {
+        return ui16(data.at(pos + 1), data.at(pos));
+    }
 
     inline qint16 i16(quint8 c1, quint8 c2) const {
         return qint32(ui16(c1, c2));
     }
+    inline qint16 i16(const QByteArray &data, qint32 pos) const {
+        return i16(data.at(pos + 1), data.at(pos));
+    }
 
     inline quint32 ui32(quint8 c1, quint8 c2, quint8 c3, quint8 c4) const {
         return (quint32(c4) << 24) | (quint32(c3) << 16) | (quint32(c2) << 8) | quint32(c1);
     }
+    inline quint32 ui32(const QByteArray &data, qint32 pos) const {
+        return ui32(data.at(pos + 3), data.at(pos + 2), data.at(pos + 1), data.at(pos));
+    }
 
     inline qint32 i32(quint8 c1, quint8 c2, quint8 c3, quint8 c4) const {
         return qint32(ui32(c1, c2, c3, c4));
     }
+    inline qint32 i32(const QByteArray &data, qint32 pos) const {
+        return i32(data.at(pos + 3), data.at(pos + 2), data.at(pos + 1), data.at(pos));
+    }
 
     static ChunkList innerFromDevice(QIODevice *d, bool *ok, IFFChunk *parent = nullptr);
 
@@ -331,6 +366,48 @@ private:
     qint32 _recursionCnt;
 };
 
+/*!
+ * \brief The IPALChunk class
+ * Interface for additional per-line palette.
+ */
+class IPALChunk : public IFFChunk
+{
+public:
+    virtual ~IPALChunk() override {}
+    IPALChunk() : IFFChunk() {}
+    IPALChunk(const IPALChunk& other) = default;
+    IPALChunk& operator =(const IPALChunk& other) = default;
+
+    /*!
+     * \brief hasAlpha
+     * \return True it the palette supports the alpha channel.
+     */
+    virtual bool hasAlpha() const { return false; }
+
+    /*!
+     * \brief clone
+     * \return A new instance of the class with all data.
+     */
+    virtual IPALChunk *clone() const = 0;
+
+    /*!
+     * \brief palette
+     * \param y The scanline.
+     * \return The modified palette.
+     */
+    virtual QList<QRgb> palette(qint32 y) const = 0;
+
+    /*!
+     * \brief initialize
+     * Initialize the palette changer.
+     * \param cmapPalette The palette as stored in the CMAP chunk.
+     * \param height The image height.
+     * \return True on success, otherwise false.
+     */
+    virtual bool initialize(const QList<QRgb>& cmapPalette, qint32 height) = 0;
+};
+
+
 /*!
  * \brief The BMHDChunk class
  * Bitmap Header
@@ -340,11 +417,22 @@ class BMHDChunk: public IFFChunk
 public:
     enum Compression {
         Uncompressed = 0, /**< Image data are uncompressed. */
-        Rle = 1 /**< Image data are RLE compressed. */
+        Rle = 1, /**< Image data are RLE compressed. */
+        RgbN8 = 4 /**< RGB8/RGBN compresson. */
     };
     enum Masking {
         None = 0, /**< Designates an opaque rectangular image. */
-        HasMask = 1, /**< A mask plane is interleaved with the bitplanes in the BODY chunk. */
+        HasMask = 1, /**< A "mask" is an optional "plane" of data the same size (w, h) as a bitplane.
+                          It tells how to "cut out" part of the image when painting it onto another
+                          image.  "One" bits in the mask mean "copy the corresponding pixel to the
+                          destination".  "Zero" mask bits mean "leave this destination pixel alone".  In
+                          other words, "zero" bits designate transparent pixels.
+                          The rows of the different bitplanes and mask are interleaved in the file.
+                          This localizes all the information pertinent to each scan line.  It
+                          makes it much easier to transform the data while reading it to adjust the
+                          image size or depth.  It also makes it possible to scroll a big image by
+                          swapping rows directly from the file without the need for random-access to
+                          all the bitplanes. */
         HasTransparentColor = 2, /**< Pixels in the source planes matching transparentColor
                                       are to be considered “transparent”. (Actually, transparentColor
                                       isn’t a “color number” since it’s matched with numbers formed
@@ -353,7 +441,7 @@ public:
                                       one of the color registers. */
         Lasso = 3 /**< The reader may construct a mask by lassoing the image as in MacPaint.
                        To do this, put a 1 pixel border of transparentColor around the image rectangle.
-                        Then do a seed fill from this border. Filled pixels are to be transparent. */
+                       Then do a seed fill from this border. Filled pixels are to be transparent. */
     };
 
     virtual ~BMHDChunk() override;
@@ -467,14 +555,64 @@ public:
 
     virtual bool isValid() const override;
 
-    QList<QRgb> palette() const;
+    /*!
+     * \brief count
+     * \return The number of color in the palette.
+     */
+    virtual qint32 count() const;
+
+    /*!
+     * \brief palette
+     * \param halfbride When True, the new palette values are appended using the halfbride method.
+     * \return The color palette.
+     * \note If \a halfbride is true, the returned palette size is count() * 2.
+     */
+    QList<QRgb> palette(bool halfbride = false) const;
 
     CHUNKID_DEFINE(CMAP_CHUNK)
 
 protected:
+    virtual QList<QRgb> innerPalette() const;
+
     virtual bool innerReadStructure(QIODevice *d) override;
 };
 
+/*!
+ * \brief The CMYKChunk class
+ *
+ * This chunk would allow color specification in terms of Cyan,
+ * Magenta, Yellow, and Black as opposed to the current CMAP which uses RGB.
+ * The format would be the same as the CMAP chunk with the exception that this
+ * chunk uses four color components as opposed to three. The number of colors
+ * contained within would be chunk length/4.  This chunk would be used in addition
+ * to the CMAP chunk.
+ */
+class CMYKChunk : public CMAPChunk
+{
+public:
+    virtual ~CMYKChunk() override;
+    CMYKChunk();
+    CMYKChunk(const CMYKChunk& other) = default;
+    CMYKChunk& operator =(const CMYKChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    /*!
+     * \brief count
+     * \return The number of color in the palette.
+     */
+    virtual qint32 count() const override;
+
+    CHUNKID_DEFINE(CMYK_CHUNK)
+
+protected:
+    /*!
+     * \brief palette
+     * \return The CMYK color palette converted to RGB one.
+     */
+    virtual QList<QRgb> innerPalette() const override;
+};
+
 /*!
  * \brief The CAMGChunk class
  */
@@ -523,13 +661,13 @@ public:
      * \brief dpiX
      * \return The horizontal resolution in DPI.
      */
-    quint16 dpiX() const;
+    virtual quint16 dpiX() const;
 
     /*!
      * \brief dpiY
      * \return The vertical resolution in DPI.
      */
-    quint16 dpiY() const;
+    virtual quint16 dpiY() const;
 
     /*!
      * \brief dotsPerMeterX
@@ -549,6 +687,50 @@ protected:
     virtual bool innerReadStructure(QIODevice *d) override;
 };
 
+/*!
+ * \brief The XBMIChunk class
+ */
+class XBMIChunk : public DPIChunk
+{
+public:
+    enum PictureType : quint16 {
+        Indexed = 0,
+        Grayscale = 1,
+        Rgb = 2,
+        RgbA = 3,
+        Cmyk = 4,
+        CmykA = 5,
+        Bitmap = 6
+    };
+
+    virtual ~XBMIChunk() override;
+    XBMIChunk();
+    XBMIChunk(const XBMIChunk& other) = default;
+    XBMIChunk& operator =(const XBMIChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    /*!
+     * \brief dpiX
+     * \return The horizontal resolution in DPI.
+     */
+    virtual quint16 dpiX() const override;
+
+    /*!
+     * \brief dpiY
+     * \return The vertical resolution in DPI.
+     */
+    virtual quint16 dpiY() const override;
+
+    /*!
+     * \brief pictureType
+     * \return The picture type
+     */
+    PictureType pictureType() const;
+
+    CHUNKID_DEFINE(XBMI_CHUNK)
+};
+
 
 /*!
  * \brief The BODYChunk class
@@ -569,13 +751,20 @@ public:
      * \brief readStride
      * \param d The device.
      * \param header The bitmap header.
+     * \param y The current scanline.
      * \param camg The CAMG chunk (optional)
      * \param cmap The CMAP chunk (optional)
-     * \param isPbm Set to true if the formType() == "PBM "
+     * \param formType The type of the current form chunk.
      * \return The scanline as requested for QImage.
      * \warning Call resetStrideRead() once before this one.
      */
-    virtual QByteArray strideRead(QIODevice *d, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, bool isPbm = false) const;
+    virtual QByteArray strideRead(QIODevice *d,
+                                  qint32 y,
+                                  const BMHDChunk *header,
+                                  const CAMGChunk *camg = nullptr,
+                                  const CMAPChunk *cmap = nullptr,
+                                  const IPALChunk *ipal = nullptr,
+                                  const QByteArray& formType = ILBM_FORM_TYPE) const;
 
     /*!
      * \brief resetStrideRead
@@ -589,15 +778,29 @@ public:
      */
     virtual bool resetStrideRead(QIODevice *d) const;
 
+    /*!
+     * \brief safeModeId
+     * \param header The header.
+     * \param camg The CAMG chunk.
+     * \return The most likely ModeId if not explicitly specified.
+     */
+    static CAMGChunk::ModeIds safeModeId(const BMHDChunk *header, const CAMGChunk *camg, const CMAPChunk *cmap = nullptr);
+
 protected:
     /*!
      * \brief strideSize
-     * \param isPbm Set true if the image is PBM.
+     * \param formType The type of the current form chunk.
      * \return The size of data to have to decode an image row.
      */
-    quint32 strideSize(const BMHDChunk *header, bool isPbm) const;
+    quint32 strideSize(const BMHDChunk *header, const QByteArray& formType) const;
 
-    QByteArray deinterleave(const QByteArray &planes, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, bool isPbm = false) const;
+    QByteArray deinterleave(const QByteArray &planes, qint32 y, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, const IPALChunk *ipal = nullptr) const;
+
+    QByteArray pbm(const QByteArray &planes, qint32 y, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, const IPALChunk *ipal = nullptr) const;
+
+    QByteArray rgb8(const QByteArray &planes, qint32 y, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, const IPALChunk *ipal = nullptr) const;
+
+    QByteArray rgbN(const QByteArray &planes, qint32 y, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, const IPALChunk *ipal = nullptr) const;
 
 private:
     mutable QByteArray _readBuffer;
@@ -619,19 +822,76 @@ public:
 
     CHUNKID_DEFINE(ABIT_CHUNK)
 
-    virtual QByteArray strideRead(QIODevice *d, const BMHDChunk *header, const CAMGChunk *camg = nullptr, const CMAPChunk *cmap = nullptr, bool isPbm = false) const override;
+    virtual QByteArray strideRead(QIODevice *d,
+                                  qint32 y,
+                                  const BMHDChunk *header,
+                                  const CAMGChunk *camg = nullptr,
+                                  const CMAPChunk *cmap = nullptr,
+                                  const IPALChunk *ipal = nullptr,
+                                  const QByteArray& formType = ACBM_FORM_TYPE) const override;
 
     virtual bool resetStrideRead(QIODevice *d) const override;
-
-private:
-    mutable qint32 _y;
 };
 
+/*!
+ * \brief The IFOR_Chunk class
+ * Interface for FORM chunks.
+ */
+class IFOR_Chunk : public IFFChunk
+{
+public:
+    virtual ~IFOR_Chunk() override;
+    IFOR_Chunk();
+
+    /*!
+     * \brief isSupported
+     * \return True if the form is supported by the plugin.
+     */
+    virtual bool isSupported() const = 0;
+
+    /*!
+     * \brief formType
+     * \return The type of image data contained in the form.
+     */
+    virtual QByteArray formType() const = 0;
+
+    /*!
+     * \brief format
+     * \return The Qt image format the form is converted to.
+     */
+    virtual QImage::Format format() const = 0;
+
+    /*!
+     * \brief transformation
+     * \return The image transformation.
+     * \note The Default implentation returns the trasformation of EXIF chunk (if any).
+     */
+    virtual QImageIOHandler::Transformation transformation() const;
+
+    /*!
+     * \brief size
+     * \return The image size in pixels.
+     */
+    virtual QSize size() const = 0;
+
+    /*!
+     * \brief optionformat
+     * \return The format retuned by the plugin after all conversions.
+     */
+    QImage::Format optionformat() const;
+
+    /*!
+     * \brief searchIPal
+     * Search the palett per line chunk.
+     * \return The per line palette (BEAM, CTBL, SHAM, etc....).
+     */
+    const IPALChunk *searchIPal() const;
+};
 
 /*!
  * \brief The FORMChunk class
  */
-class FORMChunk : public IFFChunk
+class FORMChunk : public IFOR_Chunk
 {
     QByteArray _type;
 
@@ -643,13 +903,13 @@ public:
 
     virtual bool isValid() const override;
 
-    bool isSupported() const;
+    virtual bool isSupported() const override;
 
-    QByteArray formType() const;
+    virtual QByteArray formType() const override;
 
-    QImage::Format format() const;
+    virtual QImage::Format format() const override;
 
-    QSize size() const;
+    virtual QSize size() const override;
 
     CHUNKID_DEFINE(FORM_CHUNK)
 
@@ -661,7 +921,7 @@ protected:
 /*!
  * \brief The FOR4Chunk class
  */
-class FOR4Chunk : public IFFChunk
+class FOR4Chunk : public IFOR_Chunk
 {
     QByteArray _type;
 
@@ -675,13 +935,13 @@ public:
 
     virtual qint32 alignBytes() const override;
 
-    bool isSupported() const;
+    virtual bool isSupported() const override;
 
-    QByteArray formType() const;
+    virtual QByteArray formType() const override;
 
-    QImage::Format format() const;
+    virtual QImage::Format format() const override;
 
-    QSize size() const;
+    virtual QSize size() const override;
 
     CHUNKID_DEFINE(FOR4_CHUNK)
 
@@ -689,6 +949,29 @@ protected:
     virtual bool innerReadStructure(QIODevice *d) override;
 };
 
+/*!
+ * \brief The CATChunk class
+ */
+class CATChunk : public IFFChunk
+{
+    QByteArray _type;
+
+public:
+    virtual ~CATChunk() override;
+    CATChunk();
+    CATChunk(const CATChunk& other) = default;
+    CATChunk& operator =(const CATChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    QByteArray catType() const;
+
+    CHUNKID_DEFINE(CAT__CHUNK)
+
+protected:
+    virtual bool innerReadStructure(QIODevice *d) override;
+};
+
 /*!
  * \brief The TBHDChunk class
  */
@@ -959,6 +1242,28 @@ protected:
     virtual bool innerReadStructure(QIODevice *d) override;
 };
 
+
+/*!
+ * \brief The NAMEChunk class
+ */
+class ICCNChunk : public IFFChunk
+{
+public:
+    virtual ~ICCNChunk() override;
+    ICCNChunk();
+    ICCNChunk(const ICCNChunk& other) = default;
+    ICCNChunk& operator =(const ICCNChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    QString value() const;
+
+    CHUNKID_DEFINE(ICCN_CHUNK)
+
+protected:
+    virtual bool innerReadStructure(QIODevice *d) override;
+};
+
 /*!
  * \brief The ICCPChunk class
  */
@@ -1089,4 +1394,175 @@ protected:
     virtual bool innerReadStructure(QIODevice *d) override;
 };
 
+
+/*!
+ * *** UNDOCUMENTED CHUNKS ***
+ */
+
+/*!
+ * \brief The BEAMChunk class
+ */
+class BEAMChunk : public IPALChunk
+{
+public:
+    virtual ~BEAMChunk() override;
+    BEAMChunk();
+    BEAMChunk(const BEAMChunk& other) = default;
+    BEAMChunk& operator =(const BEAMChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    virtual IPALChunk *clone() const override;
+
+    virtual QList<QRgb> palette(qint32 y) const override;
+
+    virtual bool initialize(const QList<QRgb>& cmapPalette, qint32 height) override;
+
+    CHUNKID_DEFINE(BEAM_CHUNK)
+
+protected:
+    virtual bool innerReadStructure(QIODevice *d) override;
+
+private:
+    qint32 _height;
+};
+
+/*!
+ * \brief The CTBLChunk class
+ */
+class CTBLChunk : public BEAMChunk
+{
+public:
+    virtual ~CTBLChunk() override;
+    CTBLChunk();
+    CTBLChunk(const CTBLChunk& other) = default;
+    CTBLChunk& operator =(const CTBLChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    CHUNKID_DEFINE(CTBL_CHUNK)
+};
+
+/*!
+ * \brief The SHAMChunk class
+ */
+class SHAMChunk : public IPALChunk
+{
+public:
+    virtual ~SHAMChunk() override;
+    SHAMChunk();
+    SHAMChunk(const SHAMChunk& other) = default;
+    SHAMChunk& operator =(const SHAMChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    virtual IPALChunk *clone() const override;
+
+    virtual QList<QRgb> palette(qint32 y) const override;
+
+    virtual bool initialize(const QList<QRgb>& cmapPalette, qint32 height) override;
+
+    CHUNKID_DEFINE(SHAM_CHUNK)
+
+protected:
+    virtual bool innerReadStructure(QIODevice *d) override;
+
+private:
+    qint32 _height;
+};
+
+/*!
+ * \brief The RASTChunk class
+ * \note I found an Atari STE image with the RAST chunk outside
+ *       the form chunk (Fish.neo.iff). To support it the IFF parser
+ *       should be changed so, this kind of IFFs are shown wrong.
+ */
+class RASTChunk : public IPALChunk
+{
+public:
+    virtual ~RASTChunk() override;
+    RASTChunk();
+    RASTChunk(const RASTChunk& other) = default;
+    RASTChunk& operator =(const RASTChunk& other) = default;
+
+    virtual bool isValid() const override;
+
+    virtual IPALChunk *clone() const override;
+
+    virtual QList<QRgb> palette(qint32 y) const override;
+
+    virtual bool initialize(const QList<QRgb>& cmapPalette, qint32 height) override;
+
+    CHUNKID_DEFINE(RAST_CHUNK)
+
+protected:
+    virtual bool innerReadStructure(QIODevice *d) override;
+
+private:
+    qint32 _height;
+};
+
+/*!
+ * \brief The PCHGChunk class
+ */
+class PCHGChunk : public IPALChunk
+{
+public:
+    enum Compression {
+        Uncompressed,
+        Huffman
+    };
+
+    enum Flag {
+        None = 0x00,
+        F12Bit = 0x01,
+        F32Bit = 0x02,
+        UseAlpha = 0x04
+    };
+    Q_DECLARE_FLAGS(Flags, Flag)
+
+    virtual ~PCHGChunk() override;
+    PCHGChunk();
+    PCHGChunk(const PCHGChunk& other) = default;
+    PCHGChunk& operator =(const PCHGChunk& other) = default;
+
+    Compression compression() const;
+
+    Flags flags() const;
+
+    qint16 startLine() const;
+
+    quint16 lineCount() const;
+
+    quint16 changedLines() const;
+
+    quint16 minReg() const;
+
+    quint16 maxReg() const;
+
+    quint16 maxChanges() const;
+
+    quint32 totalChanges() const;
+
+    virtual bool hasAlpha() const override;
+
+    virtual bool isValid() const override;
+
+    virtual IPALChunk *clone() const override;
+
+    virtual QList<QRgb> palette(qint32 y) const override;
+
+    virtual bool initialize(const QList<QRgb>& cmapPalette, qint32 height) override;
+
+    CHUNKID_DEFINE(PCHG_CHUNK)
+
+protected:
+    virtual bool innerReadStructure(QIODevice *d) override;
+
+private:
+    QHash<qint32, QHash<quint16, QRgb>> _paletteChanges;
+
+    QHash<qint32, QList<QRgb>> _palettes;
+};
+
 #endif // KIMG_CHUNKS_P_H
